Public Class MultiClientTcpListener

    Private _PortNum As Integer

    Public _ClientSockets As New List(Of ClientHandler)

    Private _TcpListener As System.Net.Sockets.TcpListener

    Private _ContinueAutoRemove As Boolean
    Private _ThreadAutoRemove As Threading.Thread

    Private _ContinueListenNew As Boolean
    Private _ThreadListen As Threading.Thread

    Event ClientOpen(ByRef sender As ClientHandler)
    Event ClientClose(ByRef sender As ClientHandler)
    Event ClientReceive(ByRef sender As ClientHandler, ByVal Text As String)
    Event ClientReceiveLine(ByRef sender As ClientHandler, ByVal Text As String)

    Public Sub New(ByVal PortNum As Integer)
        _PortNum = PortNum
    End Sub

    Public Sub DoStart()
        _ClientSockets.Clear()

        ' suppression automatique des clients ferms
        _ContinueAutoRemove = True
        _ThreadAutoRemove = New Threading.Thread(New Threading.ThreadStart(AddressOf AutoRemove))
        _ThreadAutoRemove.Start()

        ' coute de nouveaux clients
        _ContinueListenNew = True
        _ThreadListen = New Threading.Thread(New Threading.ThreadStart(AddressOf ListenNew))
        _ThreadListen.Start()

    End Sub

    Public Sub DoStop()

        ' fermeture de la socket d'coute
        _TcpListener.Stop()

        ' fermeture des clients
        For Each x As ClientHandler In Me._ClientSockets
            x.DoStop()
        Next

        ' suppression automatique des clients ferms
        _ContinueAutoRemove = False
        _ThreadAutoRemove.Abort()

        ' coute de nouveaux clients
        _ContinueListenNew = False
        _ThreadListen.Abort()

    End Sub

    Private Sub AutoRemove()
        Dim TmpList As List(Of ClientHandler)
        While _ContinueAutoRemove
            TmpList = New List(Of ClientHandler)(_ClientSockets)
            For Each x As ClientHandler In TmpList
                If Not x.IsAlive Then
                    RaiseEvent ClientClose(x)
                    _ClientSockets.Remove(x)
                End If
            Next
            Threading.Thread.Sleep(200)
        End While
    End Sub

    Private Sub ListenNew()
        _TcpListener = New System.Net.Sockets.TcpListener(New System.Net.IPEndPoint(0, _PortNum))
        Try
            _TcpListener.Start()
            ' Start listening for connections.
            While _ContinueListenNew
                Dim handler As System.Net.Sockets.TcpClient = _TcpListener.AcceptTcpClient()
                If Not (handler Is Nothing) Then
                    Dim NewClient As New ClientHandler(handler, Me)
                    _ClientSockets.Add(NewClient)
                    NewClient.DoStart()
                    RaiseEvent ClientOpen(NewClient)
                End If
            End While
        Catch e As Exception
        End Try
    End Sub

    Public Sub _ClientRecv(ByVal sender As ClientHandler, ByVal Text As String)
        RaiseEvent ClientReceive(sender, Text)
    End Sub

    Public Sub _ClientRecvLine(ByVal sender As ClientHandler, ByVal Text As String)
        RaiseEvent ClientReceiveLine(sender, Text)
    End Sub

    Public Class ClientHandler

        Private _ClientSocket As System.Net.Sockets.TcpClient
        Private _Stream As System.Net.Sockets.NetworkStream
        Private _ContinueProcess As Boolean = False
        Private _ClientThread As Threading.Thread
        Private _RecvBuf As String = ""
        Private _Father As MultiClientTcpListener

        Public Sub New(ByVal ClientSocket As System.Net.Sockets.TcpClient, ByRef Father As MultiClientTcpListener)
            _ClientSocket = ClientSocket
            _Father = Father
        End Sub

        Public Sub DoStart()
            If Not (_ClientSocket Is Nothing) Then
                _Stream = _ClientSocket.GetStream()
            End If
            _ContinueProcess = True
            _ClientThread = New Threading.Thread(New Threading.ThreadStart(AddressOf Process))
            _ClientThread.Start()
        End Sub

        Private Sub Process()
            ' Data buffer for incoming data.
            Dim bytes As Byte()
            If Not (_ClientSocket Is Nothing) Then
                _ClientSocket.ReceiveTimeout = 100  ' 1000 miliseconds
                While _ContinueProcess
                    ReDim bytes(_ClientSocket.ReceiveBufferSize)
                    Try
                        Dim BytesRead As Integer = _Stream.Read(bytes, 0, _ClientSocket.ReceiveBufferSize)
                        If BytesRead > 0 Then
                            Dim data As String = System.Text.Encoding.ASCII.GetString(bytes, 0, BytesRead)
                            _Receive(data)
                        End If
                    Catch e As System.IO.IOException ' Timeout
                    Catch e As System.Net.Sockets.SocketException
                        ''break()
                    End Try
                    Threading.Thread.Sleep(200)
                End While
                '_ClientSocket.Close()
            End If
        End Sub

        Public Sub DoStop()
            _ContinueProcess = False
            _ClientSocket.Close()
            'If Not (_ClientThread Is Nothing) And _ClientThread.IsAlive Then _ClientThread.Join()
        End Sub

        Public ReadOnly Property IsAlive() As Boolean
            Get
                Return _ContinueProcess
                'If Not _ContinueProcess Then Return False
                'Return Not (_ClientThread Is Nothing) And _ClientThread.IsAlive
            End Get
        End Property

        Public Sub Write(ByVal Txt As String)
            Dim sendBytes As Byte() = System.Text.Encoding.ASCII.GetBytes(Txt)
            _Stream.Write(sendBytes, 0, sendBytes.Length)
        End Sub

        Public Sub WriteLine(ByVal Txt As String)
            Write(Txt & vbCrLf)
        End Sub

        Private Sub _Receive(ByVal Txt As String)
            If Txt = "" Then Return
            Receive(Txt)
            _RecvBuf &= Txt
            _RecvBuf = _RecvBuf.Replace(vbCrLf, vbCr) ' .Replace(vbLf, vbCr)

            Dim _RecvBuf2 As String = ""
            For Each c As String In _RecvBuf.ToCharArray
                If c = vbBack Then
                    If Not _RecvBuf2.EndsWith(vbCr) And _RecvBuf2 <> "" Then
                        _RecvBuf2 = _RecvBuf2.Substring(0, _RecvBuf2.Length - 1)
                    End If
                Else
                    _RecvBuf2 &= c
                End If
            Next
            _RecvBuf = _RecvBuf2

            While _RecvBuf.Contains(vbCr)
                ReceiveLine(_RecvBuf.Substring(0, _RecvBuf.IndexOf(vbCr)))
                _RecvBuf = _RecvBuf.Substring(_RecvBuf.IndexOf(vbCr) + 1)
            End While

        End Sub

        Private Sub Receive(ByVal Txt As String)
            _Father._ClientRecv(Me, Txt)
        End Sub

        Private Sub ReceiveLine(ByVal Txt As String)
            _Father._ClientRecvLine(Me, Txt)
        End Sub

    End Class

End Class
